/*
 *  FUSE implementation of chunkfs techniques using libext2fs
 *
 *  Copyright (C) 2006 Amit Gud <gud@ksu.edu>
 *
 *  This program can be distributed under the terms of the GNU GPL v2.
 *  See the file COPYING.
 */

#define FUSE_USE_VERSION 25

#include <fuse_lowlevel.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>

#include "chunkfs.h"

/* chunk disks should be in the format chunk0.dsk, chunk1.dsk, ... */
#define NR_CHUNKS	1
#define CHUNKS		"/chunkfs/disks"

#define PATH_LEN	64

chunk_filsys fs[NR_CHUNKS];
int in_use;		/* how many chunks are actually in use */

int chunkfs_dirent_size(int len)
{
	return(fuse_dirent_size(len));
}

char *chunkfs_add_dirent(char *buf, const char *name, const struct stat *stbuf, off_t off)
{
	return(fuse_add_dirent(buf, name, stbuf, off));
}

void op_getattr(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi)
{
	int rc;
	struct stat stbuf;
	chunkfs_inode_t inode;

	memset(&stbuf, 0, sizeof(stbuf));
	rc = read_inode(fs[0], CHUNKFS_INO(ino), &inode);
	if (rc) {
		fuse_reply_err(req, ENOENT);
		return;
	}

	fill_statbuf(ino, &inode, &stbuf);
	fuse_reply_attr(req, &stbuf, 1.0);
}

static void op_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
				int to_set, struct fuse_file_info *fi)
{
	int rc;
	struct stat stbuf;
	chunkfs_inode_t inode;

	memset(&stbuf, 0, sizeof(stbuf));
	rc = read_inode(fs[0], CHUNKFS_INO(ino), &inode);
	if(rc) {
		fuse_reply_err(req, ENOENT);
		return;
	}

	if(to_set & FUSE_SET_ATTR_MODE)
		inode.i_mode = attr->st_mode;

	if(to_set & FUSE_SET_ATTR_UID)
		inode.i_uid = attr->st_uid;

	if(to_set & FUSE_SET_ATTR_GID)
		inode.i_gid = attr->st_gid;

	if(to_set & FUSE_SET_ATTR_SIZE)
		inode.i_size = attr->st_size;

	if(to_set & FUSE_SET_ATTR_ATIME)
		inode.i_atime = attr->st_atime;

	if(to_set & FUSE_SET_ATTR_MTIME)
		inode.i_mtime = attr->st_mtime;

	inode.i_mtime = time(NULL);

	rc = write_inode(fs[0], CHUNKFS_INO(ino), &inode);
	if(rc) {
		fuse_reply_err(req, EIO);
		return;
	}

	fill_statbuf(ino, &inode, &stbuf);
	fuse_reply_attr(req, &stbuf, 1.0);
}

static void op_lookup(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	struct fuse_entry_param fe;
	chunkfs_inode_t inode;
	chunkfs_ino_t ino;
	int rc;

	if(!strncmp(name, "(null)", 6)) {
		fuse_reply_err(req, ENOENT);
		return;
	}

	rc = do_lookup(fs[0], CHUNKFS_INO(parent), name, &ino, &inode);
	if (rc) {
		fuse_reply_err(req, rc);
		return;
	}

	fe.ino = ino;
	fe.generation = inode.i_generation;
	fill_statbuf(ino, &inode, &fe.attr);
	fe.attr_timeout = 2.0;
	fe.entry_timeout = 2.0;
	fuse_reply_entry(req, &fe);
}

static int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
                             off_t off, size_t maxsize)
{
	if (off < bufsize)
		return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
	else
		return fuse_reply_buf(req, NULL, 0);
}

static void op_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
                             off_t off, struct fuse_file_info *fi)
{
	errcode_t rc;
	struct dirbuf b;

	memset(&b, 0, sizeof(b));

	rc = do_dir_iterate(fs[0], CHUNKFS_INO(ino), 0, &b);
	if (rc) {
		fuse_reply_err(req, EIO);
		free(b.p);
		return;
	}

	reply_buf_limited(req, b.p, b.size, off, size);
	free(b.p);
}

static void op_open(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi)
{
	chunkfs_file_t efile = do_open(fs[0], ino, fi->flags);

	if(!efile)
		fuse_reply_err(req, EIO);
	else {
		fi->fh = (unsigned long) efile;
		fuse_reply_open(req, fi);
	}
}

static void op_create(fuse_req_t req, fuse_ino_t parent, const char *name,
		mode_t mode, struct fuse_file_info *fi)
{
	int rc;
	chunkfs_ino_t ino;
	chunkfs_file_t efile;
	chunkfs_inode_t inode;
	struct fuse_entry_param fep;

	/* write the inode */
	rc = do_create(fs[0], CHUNKFS_INO(parent), name, mode, &ino, &inode);
	if(rc)
		goto err_out;

	/* open the file */
	efile = do_open(fs[0], ino, fi->flags & ~O_CREAT);
	if(!efile) {
		rc = EIO;
		goto err_out;
	}

	fi->fh = (unsigned long) efile;
	fep.ino = ino;
	fill_statbuf(ino, &inode, &fep.attr);
	fuse_reply_create(req, &fep, fi);
	return;

err_out:
	fuse_reply_err(req, rc);
}

static void op_unlink(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	int ret;

	ret = do_unlink(fs[0], CHUNKFS_INO(parent), name, -1);
	fuse_reply_err(req, ret);
}

static void op_move(fuse_req_t req, fuse_ino_t parent, const char *name,
			fuse_ino_t newparent, const char *newname)
{
	int ret;
	chunkfs_ino_t ino;

	/* first link, then unlink */
	ret = get_ino_by_name(fs[0], CHUNKFS_INO(parent), name, &ino);
	if(ret) {
		fuse_reply_err(req, ENOENT);
		return;
	}

	ret = do_link(fs[0], CHUNKFS_INO(newparent), newname, ino, EXT2_FT_REG_FILE);
	if(ret) {
		fuse_reply_err(req, ENOENT);
		return;
	}

	ret = do_unlink(fs[0], CHUNKFS_INO(parent), name, 0);
	fuse_reply_err(req, ret);
}

static void op_read(fuse_req_t req, fuse_ino_t ino, size_t size,
			off_t off, struct fuse_file_info *fi)
{
	int rc;
	void *buf = malloc(size);
	unsigned int bytes;

	rc = do_read(CHUNKFS_FILE(fi->fh), CHUNKFS_INO(ino), size, off, buf, &bytes);
	if(rc)
		fuse_reply_err(req, rc);
	else
		fuse_reply_buf(req, buf, bytes);

	free(buf);
}

static void op_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
			size_t size, off_t off, struct fuse_file_info *fi)
{
	int rc;
	unsigned int bytes;

	rc = do_write(CHUNKFS_FILE(fi->fh), ino, buf, size, off, &bytes);
	if(rc)
		fuse_reply_err(req, rc);
	else
		fuse_reply_write(req, bytes);
}

static void op_flush(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi)
{
	int rc;
	rc = do_file_flush(CHUNKFS_FILE(fi->fh));
	fuse_reply_err(req, rc);
}

static void op_release(fuse_req_t req, fuse_ino_t ino,
			struct fuse_file_info *fi)
{
	int rc;
	rc = do_file_close(CHUNKFS_FILE(fi->fh));
	if (rc)
		fuse_reply_err(req, EIO);
	else
		fuse_reply_err(req, 0);
}

static void op_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
			mode_t mode)
{
	struct fuse_entry_param fe;
	chunkfs_ino_t ino;
	chunkfs_inode_t inode;
	int ret;

	if(parent == 1) parent = CHUNKFS_ROOT_INO;

	ret = do_mkdir(fs[0], parent, name, mode, &ino);
	if(ret) {
		fuse_reply_err(req, ret);
		return;
	}

	ret = read_inode(fs[0], ino, &inode);
	if(ret) {
		fuse_reply_err(req, EIO);
		return;
	}

	fe.ino = ino;
	fe.generation = inode.i_generation;
	fill_statbuf(ino, &inode, &fe.attr);
	fe.attr_timeout = 2.0;
	fe.entry_timeout = 2.0;

	fuse_reply_entry(req, &fe);
}

static void op_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	int ret;

	ret = do_rmdir(fs[0], CHUNKFS_INO(parent), name);
	fuse_reply_err(req, ret);
}

static int check_chunk(int n)
{
	errcode_t rc;
	chunkfs_ino_t ino;

	rc = ext2fs_lookup(fs[n], CHUNKFS_ROOT_INO, ".chunkfs", strlen(".chunkfs"),
			NULL, &ino);

	if(rc)
		rc = do_mkdir(fs[n], CHUNKFS_ROOT_INO, ".chunkfs", 0700, &ino);
	return rc;
}

static void op_init(void *userdata)
{
	errcode_t ret;
	int i;
	io_channel data_io = 0;
	char *device;

	device = (char *) malloc(strlen(CHUNKS) + 10);

	if(!device) {
		printf("chunkfs: malloc error\n");
		exit(1);
	}

	in_use = 0;
	for(i = 0; i < NR_CHUNKS; i++) {
		sprintf(device, CHUNKS"/chunk%d.dsk", i);

		ret = ext2fs_open(device, EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK, 0,
					0, unix_io_manager, &fs[in_use]);
		if(ret) {
			com_err("chunkfs", ret, "while trying to open %s", device);
			goto close_cont;
		}

		ret = ext2fs_read_inode_bitmap(fs[in_use]);
		if(ret) {
			com_err(device, ret, "while reading inode bitmap");
			goto close_cont;
		}

		ret = ext2fs_read_block_bitmap(fs[in_use]);
		if(ret) {
			com_err(device, ret, "while reading block bitmap");
			goto close_cont;
		}

		/*if(check_chunk(in_use)) {
			printf("chunkfs: chunk %s not initialized or not empty\n", device);
			goto close_cont;
		}*/

		in_use++;
		continue;

close_cont:
		ext2fs_close(fs[in_use]);
	}

	printf("chunkfs initialized with %d chunks\n", in_use);
	free(device);
}

static void op_destroy(void *userdata)
{
	errcode_t ret;
	int i;

	for(i = 0; i < in_use; i++) {
		ret = ext2fs_close(fs[i]);
		if(ret)
			com_err("chunkfs", ret, "while trying to close chunk %d", i);
	}
	printf("chunkfs destroyed\n");
}

static struct fuse_lowlevel_ops chunkfs_ops = {
	.init           = op_init,
	.destroy        = op_destroy,
	.lookup         = op_lookup,
	.forget         = NULL,
	.getattr        = op_getattr,
	.setattr        = op_setattr,
	.readlink       = NULL,
	.mknod          = NULL,
	.mkdir          = op_mkdir,
	.rmdir          = op_rmdir,
	.symlink        = NULL,
	.rename         = op_move,
	.link           = NULL,
	.unlink         = op_unlink,
	.open           = op_open,
	.read           = op_read,
	.write          = op_write,
	.flush          = op_flush,
	.release        = op_release,
	.fsync          = NULL,
	.opendir        = op_open,
	.readdir        = op_readdir,
	.releasedir     = NULL,
	.fsyncdir       = NULL,
	.statfs         = NULL,
	.setxattr       = NULL,
	.getxattr       = NULL,
	.listxattr      = NULL,
	.removexattr    = NULL,
	.access         = NULL,
	.create         = op_create,
};

int main(int argc, char *argv[])
{
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	char *mountpoint;
	int err = -1;
	int fd;

	if (fuse_parse_cmdline(&args, &mountpoint, NULL, NULL) != -1 &&
		(fd = fuse_mount(mountpoint, &args)) != -1) {
		struct fuse_session *se;

		se = fuse_lowlevel_new(&args, &chunkfs_ops, sizeof(chunkfs_ops),
				NULL);
		if (se != NULL) {
			if (fuse_set_signal_handlers(se) != -1) {
				struct fuse_chan *ch = fuse_kern_chan_new(fd);
				if (ch != NULL) {
					fuse_session_add_chan(se, ch);
					err = fuse_session_loop(se);
				}
				fuse_remove_signal_handlers(se);
			}
			fuse_session_destroy(se);
		}
		close(fd);
	}
	fuse_unmount(mountpoint);
	fuse_opt_free_args(&args);

	return err ? 1 : 0;
}

